import argparse
import sys
from typing import Any
from typing import Dict
from typing import Iterable

from rich.console import Console
from safdie import SafdieRunner

from .constants import COMMAND_ENTRYPOINT_NAME
from .exceptions import DejimaError
from .exceptions import DejimaUserError
from .plugin import CommandPlugin


class Runner(SafdieRunner):
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("--debugger", action="store_true")
        return super().add_arguments(parser)

    def handle(
        self,
        args: argparse.Namespace,
        init_args: Iterable[Any],
        init_kwargs: Dict[str, Any],
        handle_args: Iterable[Any],
        handle_kwargs: Dict[str, Any],
    ) -> Any:
        console = Console()

        if args.debugger:
            import debugpy

            debugpy.listen(("0.0.0.0", 5678))
            console.print("[red][italic]Awaiting debugger connection...[/italic][/red]")
            debugpy.wait_for_client()

        init_kwargs["console"] = console

        return super().handle(args, init_args, init_kwargs, handle_args, handle_kwargs)


def main(argv=sys.argv):
    console = Console()

    try:
        Runner(
            COMMAND_ENTRYPOINT_NAME,
            CommandPlugin,
        ).run()
    except DejimaError as e:
        console.print(f"[red]{e}[/red]")
    except DejimaUserError as e:
        console.print(f"[yellow]{e}[/yellow]")
    except Exception:
        console.print_exception(show_locals=True)



import argparse
import base64
import datetime
from textwrap import dedent
from typing import List

from ..api import AnkiCardTemplate
from ..api import AnkiMediaUpload
from ..api import AnkiModel
from ..api import AnkiNote
from ..api import AnkiNoteOptions
from ..api import Connection as AnkiConnection
from ..api import escape
from ..db import Connection as DatabaseConnection
from ..plugin import CommandPlugin
from ..plugin import Note
from ..plugin import get_installed_sources


class ImportCommand(CommandPlugin):
    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        sources = get_installed_sources()

        parser.add_argument("deck_name", type=str)
        parser.add_argument("--reimport", action="store_true", default=False)
        subparsers = parser.add_subparsers(dest="source")
        subparsers.required = True

        for src_name, src_class in sources.items():
            parser_kwargs = {}

            src_help = src_class.get_help()
            if src_help:
                parser_kwargs["help"] = src_help

            subparser = subparsers.add_parser(src_name, **parser_kwargs)
            src_class.add_arguments(subparser)

        return super().add_arguments(parser)

    def handle(self) -> None:
        run_timestamp = datetime.datetime.utcnow()
        import_name = f'import{run_timestamp.strftime("%Y%m%dT%H%M%S")}'

        db = DatabaseConnection()
        api = AnkiConnection()

        sources = get_installed_sources()
        source = sources[self.options.source](
            self.options.source,
            self.options,
            self.console,
        )

        model_name = source.get_model_name()
        if model_name not in api.get_model_names():
            model = AnkiModel(
                source.get_model_name(),
                list(source.fields.keys()),
                source.get_card_style(),
                source.get_is_cloze(),
                [
                    AnkiCardTemplate(
                        Name=dedent(t.name).strip(),
                        Front=dedent(t.front).strip(),
                        Back=dedent(t.back).strip(),
                    )
                    for t in source.get_card_templates()
                ],
            )
            api.create_model(model)

        added = 0
        merged = 0
        invalid = 0
        total = 0

        for idx, (foreign_key, entry) in enumerate(source.get_entries()):
            try:
                total += 1
                if (
                    not foreign_key
                    or self.options.reimport
                    or not db.annotation_is_known(self.options.source, foreign_key)
                ):
                    note_is_valid = True
                    for field_name, field_info in source.fields.items():
                        if not field_info.optional and not entry.fields.get(field_name):
                            self.console.print(
                                f"[yellow]Invalid note (missing '{field_name}') "
                                f"[bold]Idx {idx}[/bold] "
                                "[/yellow]"
                            )
                            invalid += 1
                            note_is_valid = False
                            if foreign_key:
                                db.mark_entry_processed(
                                    self.options.source, foreign_key, None, import_name
                                )

                        if field_info.default and field_name not in entry.fields:
                            entry.fields[field_name] = field_info.default
                    if not note_is_valid:
                        continue
                    clauses: List[str] = []
                    for field_name, field_info in source.fields.items():
                        if field_info.unique:
                            clauses.append(
                                f"{field_name}:{escape(entry.fields.get(field_name, ''))}"
                            )
                            pass

                    duplicates = []
                    if clauses:
                        duplicates = api.find_notes(
                            f"""
                            deck:{escape(self.options.deck_name)}
                            (
                                {' or '.join(clauses)}
                            )
                        """
                        )

                    if duplicates:
                        if len(duplicates) > 1:
                            self.console.print(
                                "[red]Multiple duplicate notes found for "
                                f"entry {entry} (idx: {idx}): "
                                f"{duplicates}.[/red]"
                            )

                        anki_id = duplicates[0]
                        duplicate_anki = api.get_note(anki_id)

                        duplicate_note = Note(
                            fields=duplicate_anki.fields, tags=duplicate_anki.tags
                        )

                        anki_note = source.resolve_duplicate(duplicate_note, entry)

                        duplicate_anki.fields = anki_note.fields
                        duplicate_anki.tags = anki_note.tags

                        api.update_note(anki_id, duplicate_anki)
                        if foreign_key:
                            db.mark_entry_processed(
                                self.options.source, foreign_key, anki_id, import_name
                            )
                        self.console.print(
                            "[bright_green]Updated note "
                            f"[bold]{anki_id}[/bold][/bright_green]"
                        )
                        merged += 1
                    else:
                        new_note = AnkiNote(
                            model_name,
                            self.options.deck_name,
                            entry.fields,
                            entry.tags
                            + [
                                "dejima-import",
                                import_name,
                            ],
                        )
                        anki_id = api.add_note(
                            new_note, AnkiNoteOptions(allowDuplicate=True)
                        )
                        if foreign_key:
                            db.mark_entry_processed(
                                self.options.source, foreign_key, anki_id, import_name
                            )
                        added += 1
                        self.console.print(
                            "[green]Created note " f"[bold]{anki_id}[/bold]" "[/green]"
                        )

                    # We can upload the media regardless of whether
                    # this element turns out to be a duplicate --
                    # Anki will search for and find unreferenced media
                    # automatically.
                    for media in entry.media:
                        api.store_media_file(
                            AnkiMediaUpload(
                                filename=media.filename,
                                data=base64.b64encode(media.data).decode("ascii"),
                            )
                        )

            except Exception:
                self.console.print(
                    f"[red]Added [bold]{added}[/bold] new records[/red] "
                    f"({merged} merged; {invalid} invalid; "
                    f"{total - added - invalid - merged} already processed)"
                )
                self.console.print(
                    f'[red][bold]Import "{import_name}" failed.[/bold][/red]'
                )
                raise

        self.console.print(
            f"[blue]Added [bold]{added}[/bold] new records[/blue] "
            f"({merged} merged; {invalid} invalid; "
            f"{total - added - invalid - merged} already processed)"
        )



import datetime
import os.path
import sqlite3
from typing import Optional

import appdirs

from . import constants

USER_DATA_DIR = appdirs.user_data_dir(constants.APP_NAME, constants.AUTHOR_NAME)
DB_PATH = os.path.join(USER_DATA_DIR, "dejima.db")


class Connection:
    _db: sqlite3.Connection

    def __init__(self):
        if not os.path.exists(USER_DATA_DIR):
            os.makedirs(USER_DATA_DIR, exist_ok=True)

        self._db = sqlite3.Connection(
            DB_PATH,
            detect_types=sqlite3.PARSE_DECLTYPES,
            isolation_level=None,
        )

        self._create_tables()

        super().__init__()

    def get_cursor(self) -> sqlite3.Cursor:
        return self._db.cursor()

    def _create_tables(self):
        cursor = self.get_cursor()
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS known_entries (
                key string,
                source string,
                anki_id integer,
                importName string,
                imported timestamp
            )
        """
        )
        cursor.close()

    def mark_entry_processed(
        self, source: str, key: str, anki_id: Optional[int], import_name: str
    ):
        cursor = self.get_cursor()
        cursor.execute(
            """
            INSERT INTO known_entries
                (key, source, anki_id, imported, importName)
            VALUES
                (?, ?, ?, ?, ?)
        """,
            (key, source, anki_id, datetime.datetime.utcnow(), import_name),
        )
        cursor.close()

    def annotation_is_known(self, source: str, key: str) -> bool:
        cursor = self.get_cursor()
        cursor.execute(
            """
            SELECT 1
            FROM known_entries
            WHERE key = ? AND source = ?
        """,
            (
                key,
                source,
            ),
        )

        exists = len(cursor.fetchall()) > 0
        cursor.close()

        return exists



from __future__ import annotations

import argparse
import dataclasses
import logging
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from rich.console import Console
from safdie import BaseCommand
from safdie import get_entrypoints

from .constants import SOURCE_ENTRYPOINT_NAME

logger = logging.getLogger(__name__)


def get_installed_sources() -> Dict[str, Type[SourcePlugin]]:
    return get_entrypoints(SOURCE_ENTRYPOINT_NAME, SourcePlugin)


class NoteField:
    _unique: bool
    _attribute_name: str
    _field_name_override: Optional[str]
    _default: Optional[str]
    _merge: bool
    _optional: bool

    def __init__(
        self,
        field_name: Optional[str] = None,
        unique: bool = False,
        default: str = "",
        merge: bool = False,
        optional: bool = False,
    ):
        if field_name is not None and not field_name:
            raise ValueError("Specified field name is not valid.")

        self._unique = unique
        self._field_name_override = field_name
        self._default = default
        self._merge = merge
        self._optional = optional

        super().__init__()

    def _set_attribute_name(self, value: str) -> None:
        self._attribute_name = value

    @property
    def field_name(self) -> str:
        if self._field_name_override:
            return self._field_name_override

        return self._attribute_name

    @property
    def optional(self) -> bool:
        return self._optional

    @property
    def unique(self) -> bool:
        return self._unique

    @property
    def default(self) -> Optional[str]:
        return self._default

    @property
    def merge(self) -> bool:
        return self._merge


@dataclasses.dataclass
class CardTemplate:
    name: str
    front: str
    back: str


@dataclasses.dataclass
class Media:
    filename: str
    data: bytes


@dataclasses.dataclass
class Note:
    fields: Dict[str, str] = dataclasses.field(default_factory=dict)
    tags: List[str] = dataclasses.field(default_factory=list)
    media: List[Media] = dataclasses.field(default_factory=list)


class _SourcePluginBase(type):
    def __new__(cls, name, bases, namespaces, **kwargs):
        fields: Dict[str, NoteField] = {}
        field_attributes: Dict[str, str] = {}

        for attribute_name, attribute_value in namespaces.items():
            if isinstance(attribute_value, NoteField):
                attribute_value._set_attribute_name(attribute_name)

                fields[attribute_name] = attribute_value
                field_attributes[attribute_name] = attribute_value

        namespaces.update(field_attributes)
        namespaces["_fields"] = fields

        return super().__new__(cls, name, bases, namespaces, **kwargs)


class SourcePlugin(metaclass=_SourcePluginBase):
    CLOZE = False

    _entrypoint_name: str
    _options: argparse.Namespace
    _console: Console
    _fields: Dict[str, NoteField]

    def __init__(
        self, entrypoint_name: str, options: argparse.Namespace, console: Console
    ):
        self._entrypoint_name = entrypoint_name
        self._options = options
        self._console = console

        super().__init__()

    @property
    def fields(self) -> Dict[str, NoteField]:
        return {v.field_name: v for k, v in self._fields.items()}

    @property
    def console(self) -> Console:
        return self._console

    @property
    def options(self) -> argparse.Namespace:
        return self._options

    def get_model_name(self) -> str:
        return f"Dejima - {self._entrypoint_name}"

    def get_card_templates(self) -> List[CardTemplate]:
        raise NotImplementedError()

    def get_is_cloze(self) -> bool:
        return self.CLOZE

    def get_card_style(self) -> str:
        return ""

    @classmethod
    def get_help(cls) -> Optional[str]:
        return None

    @classmethod
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        pass

    def get_entries(self) -> Iterable[Tuple[Optional[str], Note]]:
        raise NotImplementedError()

    def resolve_duplicate(self, original: Note, new: Note) -> Note:
        for field_name, field_info in self.fields.items():
            if not field_info.merge:
                continue

            if original.fields.get(field_name) != new.fields.get(field_name):
                original.fields[field_name] = (
                    f"{original.fields[field_name]}\n\n<hr />\n\n"
                    f"{new.fields.get(field_name)}"
                )

        original.tags = list(set(original.tags) | set(new.tags))

        return original


class CommandPlugin(BaseCommand):
    _console: Console

    def __init__(self, options: argparse.Namespace, console: Console):
        self._console: Console = console
        super().__init__(options)

    @property
    def console(self) -> Console:
        return self._console



import shutil
import tempfile
import uuid
from unittest import TestCase
from unittest.mock import patch

from ..db import Connection


class TestDbConnection(TestCase):
    def setUp(self):
        self._tmp_dir = tempfile.mkdtemp()

        with patch("appdirs.user_data_dir") as user_data_dir:
            user_data_dir.return_value = self._tmp_dir
            self.db = Connection()

        super().setUp()

    def tearDown(self) -> None:
        shutil.rmtree(self._tmp_dir)
        return super().tearDown()

    def test_mark_entry_processed(self):
        arbitrary_source = "arbitrary source"
        arbitrary_key = str(uuid.uuid4())

        self.db.mark_entry_processed(
            arbitrary_source,
            arbitrary_key,
            10384134,
            "arbitrary import name",
        )

        actual_result = self.db.annotation_is_known(
            arbitrary_source,
            arbitrary_key,
        )
        expected_result = True

        assert actual_result == expected_result

    def test_mark_entry_processed_no_id(self):
        arbitrary_source = "arbitrary source"
        arbitrary_key = str(uuid.uuid4())

        self.db.mark_entry_processed(
            arbitrary_source,
            arbitrary_key,
            None,
            "arbitrary import name",
        )

        actual_result = self.db.annotation_is_known(
            arbitrary_source,
            arbitrary_key,
        )
        expected_result = True

        assert actual_result == expected_result

    def test_annotation_is_known_but_its_not(self):
        arbitrary_source = "arbitrary source"
        arbitrary_key = str(uuid.uuid4())

        actual_result = self.db.annotation_is_known(
            arbitrary_source,
            arbitrary_key,
        )
        expected_result = False

        assert actual_result == expected_result



import json
from unittest import TestCase
from unittest.mock import ANY
from unittest.mock import Mock
from unittest.mock import patch

import pytest
import requests

from ..api import AnkiError
from ..api import Connection


class TestApi(TestCase):
    def setUp(self):
        self.response_data = {
            "result": None,
            "error": None,
        }
        self.response = Mock(
            json=Mock(
                return_value=self.response_data,
            ),
            spec=requests.Response,
        )

        with patch("requests.Session") as session:
            session.return_value = Mock(post=Mock(return_value=self.response))

            self.session = session

            self.api = Connection()

    def test_dispatch(self):
        arbitrary_action = "some_action"
        arbitrary_params = {
            "something": "special",
        }

        self.api._dispatch(arbitrary_action, arbitrary_params)

        self.session.return_value.post.assert_called_with(
            ANY,
            data=json.dumps(
                {"action": arbitrary_action, "version": 6, "params": arbitrary_params},
                indent=4,
                sort_keys=True,
            ),
        )

    def test_dispatch_no_parms(self):
        arbitrary_action = "some_action"

        self.api._dispatch(arbitrary_action)

        self.session.return_value.post.assert_called_with(
            ANY,
            data=json.dumps(
                {"action": arbitrary_action, "version": 6},
                indent=4,
                sort_keys=True,
            ),
        )

    def test_dispatch_error(self):
        arbitrary_action = "some_action"
        arbitrary_error_msg = "some error"

        self.response_data["error"] = arbitrary_error_msg

        with pytest.raises(AnkiError):
            self.api._dispatch(arbitrary_action)



from unittest import TestCase
from unittest.mock import Mock

from .. import plugin


class TestSourcePluginMetaclass(TestCase):
    def test_basic(self):
        class MySource(plugin.SourcePlugin):
            Front = plugin.NoteField()

        source = MySource("arbitrary", Mock(), Mock())

        assert source.Front._attribute_name == "Front"
        assert source._fields



import argparse
import base64
import dataclasses
import mimetypes
import sys
import uuid
from hashlib import sha256
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from lln_json_parser import parser
from lln_json_parser import types

from ..plugin import CardTemplate
from ..plugin import Media
from ..plugin import Note
from ..plugin import NoteField
from ..plugin import SourcePlugin


@dataclasses.dataclass
class MediaDescriptor:
    filename: str
    field_value: str
    file_data: bytes


class LLNJsonSource(SourcePlugin):
    Source = NoteField()
    SourceLanguage = NoteField(optional=True)

    HumanTranslation = NoteField()
    MachineTranslation = NoteField(optional=True)
    TranslationLanguage = NoteField(optional=True)

    ThumbnailPre = NoteField(optional=True)
    ThumbnailPost = NoteField(optional=True)
    Audio = NoteField(optional=True)

    Reverse = NoteField(field_name="Add Reverse", optional=True)

    def get_card_style(self) -> str:
        return """
            .card {
                font-family: arial;
                font-size: 20px;
                text-align: center;
                color: black;
                background-color: white;
            }
        """

    def get_card_templates(self) -> List[CardTemplate]:
        return [
            CardTemplate(
                name="Source to Translation",
                front="""
                    <p>
                        {{ThumbnailPre}}
                    </p>
                    <p>
                        {{Source}}
                    </p>
                    <p>
                        {{ThumbnailPost}}
                    </p>
                    <div style="display: none;">{{Audio}}</div>
                """,
                back="""
                    <p>
                        {{FrontSide}}
                    </p>
                    <hr id='answer' />
                    <p>
                        {{HumanTranslation}}<br />
                        {{#MachineTranslation}}
                            <i>({{MachineTranslation}})</i>
                        {{/MachineTranslation}}
                    </p>
                """,
            ),
            CardTemplate(
                name="Translation to Source",
                front="""
                    {{#Add Reverse}}
                        <p>
                            {{HumanTranslation}}
                        </p>
                    {{/Add Reverse}}
                """,
                back="""
                    {{FrontSide}}
                    <hr id='answer' />
                    <p>
                        {{ThumbnailPre}}
                    </p>
                    <p>
                        {{Source}}
                    </p>
                    <p>
                        {{ThumbnailPost}}
                    </p>
                    <div style="display: none;">{{Audio}}</div>
                """,
            ),
        ]

    def _calculate_key(self, saved: Union[types.SavedPhrase, types.SavedWord]) -> str:
        data = saved.json(sort_keys=True)
        return sha256(data.encode("utf-8")).hexdigest()

    @classmethod
    def add_arguments(self, arg_parser: argparse.ArgumentParser) -> None:
        arg_parser.add_argument(
            "-i",
            "--input",
            nargs="?",
            type=argparse.FileType("r"),
            default=sys.stdin,
            help="File to read from (default: stdin)",
        )
        return super().add_arguments(arg_parser)

    def _generate_media_data(
        self, media: Union[types.Thumbnail, types.Audio]
    ) -> Optional[Tuple[str, bytes]]:
        if not media or not media.data_url:
            return None

        mime_type, _ = mimetypes.guess_type(media.data_url)
        if not mime_type:
            return None

        extension = mimetypes.guess_extension(mime_type)
        if not extension:
            return None

        filename = f"{uuid.uuid4()}{extension}"
        _, encoded_data = media.data_url.split(",")

        return filename, base64.b64decode(encoded_data)

    def _get_thumbnail_media_descriptor(
        self, thumbnail: Optional[types.Thumbnail]
    ) -> Optional[MediaDescriptor]:
        media_data = self._generate_media_data(thumbnail)
        if not media_data:
            return None

        filename, data = media_data

        return MediaDescriptor(filename, f"<img src='{filename}' />", data)

    def _get_audio_media_descriptor(
        self, audio: Optional[types.Audio]
    ) -> Optional[MediaDescriptor]:
        media_data = self._generate_media_data(audio)
        if not media_data:
            return None

        filename, data = media_data

        return MediaDescriptor(filename, f"[sound:{filename}]", data)

    def _get_media_for_phrase(
        self, phrase: types.Phrase
    ) -> Tuple[List[Media], Dict[str, str]]:
        media: List[Media] = []
        fields: Dict[str, str] = {}

        thumb_prev: Optional[MediaDescriptor] = self._get_thumbnail_media_descriptor(
            phrase.thumb_prev
        )
        if thumb_prev:
            fields[self.ThumbnailPre.field_name] = thumb_prev.field_value
            media.append(Media(thumb_prev.filename, thumb_prev.file_data))

        thumb_next: Optional[MediaDescriptor] = self._get_thumbnail_media_descriptor(
            phrase.thumb_next
        )
        if thumb_next:
            fields[self.ThumbnailPost.field_name] = thumb_next.field_value
            media.append(Media(thumb_next.filename, thumb_next.file_data))

        audio: Optional[MediaDescriptor] = self._get_audio_media_descriptor(
            phrase.audio
        )
        if audio:
            fields[self.Audio.field_name] = audio.field_value
            media.append(Media(audio.filename, audio.file_data))

        return media, fields

    def _get_note_for_saved_word(self, saved: types.SavedWord) -> Note:
        fields: Dict[str, str] = {
            self.Source.field_name: (
                f'{saved.word.text}<br /><i>"{saved.context.phrase.subtitles.target}"</i>'
            ),
            self.SourceLanguage.field_name: saved.lang_code,
            self.HumanTranslation.field_name: (
                f"{', '.join(saved.word_translations)}<br />"
                f'<i>"{saved.context.phrase.human_translations.target}"</i>'
            ),
            self.MachineTranslation.field_name: (
                saved.context.phrase.machine_translations.target
            ),
            self.TranslationLanguage.field_name: saved.translation_lang_code,
        }

        media, extra_fields = self._get_media_for_phrase(saved.context.phrase)
        fields.update(extra_fields)

        return Note(fields=fields, media=media)

    def _get_note_for_saved_phrase(self, saved: types.SavedPhrase) -> Note:
        fields: Dict[str, str] = {
            self.Source.field_name: saved.context.phrase.subtitles.target,
            self.SourceLanguage.field_name: saved.lang_code,
            self.HumanTranslation.field_name: saved.context.phrase.human_translations.target,
            self.MachineTranslation.field_name: (
                saved.context.phrase.machine_translations.target
            ),
            self.TranslationLanguage.field_name: saved.translation_lang_code,
        }

        media, extra_fields = self._get_media_for_phrase(saved.context.phrase)
        fields.update(extra_fields)

        return Note(fields=fields, media=media)

    def get_entries(self) -> Iterable[Tuple[str, Note]]:
        entries = parser.get_entries(self.options.input)

        for entry in entries:
            foreign_key = self._calculate_key(entry)

            if isinstance(entry, types.SavedPhrase):
                note = self._get_note_for_saved_phrase(entry)
            elif isinstance(entry, types.SavedWord):
                note = self._get_note_for_saved_word(entry)
            else:
                raise ValueError(f"Unexpected note type: {entry}")

            yield foreign_key, note



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")
    ) as fh:
        return fh.read()


setup(
    name="dejima",
    version="2.1.1",
    license="MIT",
    description="Easily import data from a variety of formats into Anki.",
    long_description_content_type="text/markdown",
    long_description="%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub(
            "", read("README.md")
        )
    ),
    author="Adam Coddington",
    author_email="me@adamcoddington.net",
    url="https://github.com/coddingtonbear/dejima",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        "Topic :: Utilities",
    ],
    project_urls={
        "Issue Tracker": "https://github.com/coddingtonbear/dejima/issues",
    },
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires=">=3.6",
    install_requires=[
        "requests>=2.26.0,<3.0",
        "boox-annotation-parser>=0.1.2,<1.0",
        "lln-json-parser>=0.1.3,<1.0",
        "rich>=10.7.0,<11.0",
        "appdirs>=1.4.4,<2.0",
        "safdie>=2.0.0,<3.0",
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
    },
    entry_points={
        "console_scripts": [
            "dejima = dejima.cli:main",
        ],
        "dejima.sources": [
            "boox = dejima.sources.boox:BooxSource",
            "lln-json = dejima.sources.lln:LLNJsonSource",
        ],
        "dejima.commands": ["import = dejima.commands.import:ImportCommand"],
    },
)



import dataclasses
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import requests


@dataclasses.dataclass
class AnkiMediaUpload:
    filename: str
    data: Optional[str] = None  # Base64'd bytes
    path: Optional[str] = None
    url: Optional[str] = None
    deleteExisting: bool = True


@dataclasses.dataclass
class AnkiCardTemplate:
    Name: str
    Front: str
    Back: str


@dataclasses.dataclass
class AnkiModel:
    modelName: str
    inOrderFields: List[str]
    css: str
    isCloze: bool
    cardTemplates: List[AnkiCardTemplate]


@dataclasses.dataclass
class AnkiNoteOptions:
    allowDuplicate: bool = False


@dataclasses.dataclass
class AnkiNote:
    modelName: str
    deckName: str
    fields: Dict[str, str]
    tags: List[str]


class AnkiError(Exception):
    pass


class AnkiNoteDoesNotExist(Exception):
    pass


def escape(term: str):
    replacements = {
        "\\": "\\\\",
        '"': '\\"',
        "*": "\\*",
        "_": "\\_",
        ":": "\\:",
    }
    for fr, to in replacements.items():
        term = term.replace(fr, to)

    return f'"{term}"'


class Connection:
    _hostname: str
    _port: int

    def __init__(self, hostname="127.0.0.1", port=8765):
        self._hostname = hostname
        self._port = port
        self._connection = requests.Session()

        super().__init__()

    def _dispatch(self, action: str, params: Dict[str, Any] = None) -> Any:
        payload = json.dumps(
            {
                "action": action,
                "version": 6,
                **({"params": params} if params is not None else {}),
            },
            indent=4,
            sort_keys=True,
        )
        request = self._connection.post(
            f"http://{self._hostname}:{self._port}/",
            data=payload,
        )
        request.raise_for_status()

        result = request.json()

        if result["error"] is not None:
            raise AnkiError(result["error"])

        return result["result"]

    def get_deck_names(self) -> List[str]:
        return self._dispatch("deckNames")

    def get_model_names(self) -> List[str]:
        return self._dispatch("modelNames")

    def create_model(self, model: AnkiModel) -> Dict:
        return self._dispatch("createModel", dataclasses.asdict(model))

    def add_note(self, note: AnkiNote, options: AnkiNoteOptions = None) -> int:
        note_data = dataclasses.asdict(note)
        if options is not None:
            note_data.get("options", {}).update(dataclasses.asdict(options))

        return self._dispatch("addNote", {"note": note_data})

    def update_note(self, id: int, note: AnkiNote) -> None:
        note_data = dataclasses.asdict(note)
        note_data["id"] = id

        return self._dispatch("updateNoteFields", {"note": note_data})

    def get_note(self, id) -> AnkiNote:
        notes = self._dispatch("notesInfo", {"notes": [id]})
        if len(notes) == 0:
            raise AnkiNoteDoesNotExist(id)

        result = notes[0]

        return AnkiNote(
            modelName=result.get("modelName", ""),
            deckName=result.get("deckName", ""),
            fields={k: v["value"] for k, v in result.get("fields", {}).items()},
            tags=result.get("tags", []),
        )

    def find_notes(self, query: str) -> List[int]:
        query = query.replace("\n", " ")
        return self._dispatch(
            "findNotes",
            {"query": query},
        )

    def store_media_file(self, media: AnkiMediaUpload) -> str:
        return self._dispatch("storeMediaFile", dataclasses.asdict(media))



import argparse
import dataclasses
import json
import sys
from hashlib import sha256
from typing import Iterable
from typing import List
from typing import Tuple

from boox_annotation_parser import parser as boox_parser

from ..plugin import CardTemplate
from ..plugin import Note
from ..plugin import NoteField
from ..plugin import SourcePlugin


class BooxSource(SourcePlugin):
    Front = NoteField(unique=True, merge=True)
    Back = NoteField(unique=True, merge=True)
    Reverse = NoteField(default="1", field_name="Add Reverse", optional=True)

    def get_card_style(self) -> str:
        return """
            .card {
                font-family: arial;
                font-size: 20px;
                text-align: center;
                color: black;
                background-color: white;
            }
        """

    def get_card_templates(self) -> List[CardTemplate]:
        return [
            CardTemplate(
                name="Card 1",
                front="<p>{{Front}}</p>",
                back="""
                    {{FrontSide}}
                    <hr id='answer' />
                    <p>
                        {{Back}}
                    </p>
                """,
            ),
            CardTemplate(
                name="Card 2",
                front="""
                    {{#Add Reverse}}
                        <p>
                            {{Back}}
                        </p>
                    {{/Add Reverse}}
                """,
                back="""
                    {{FrontSide}}
                    <hr id='answer' />
                    <p>
                        {{Front}}
                    </p>
                """,
            ),
        ]

    @classmethod
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-i",
            "--input",
            nargs="?",
            type=argparse.FileType("r"),
            default=sys.stdin,
            help="File to read from (default: stdin)",
        )
        return super().add_arguments(parser)

    def _calculate_key(self, note: boox_parser.Annotation) -> str:
        data = json.dumps(dataclasses.asdict(note), sort_keys=True, default=str)
        return sha256(data.encode("utf-8")).hexdigest()

    def get_entries(self) -> Iterable[Tuple[str, Note]]:
        annotation_info = boox_parser.get_annotations(self.options.input)
        for annotation in annotation_info.annotations:
            foreign_key = self._calculate_key(annotation)
            note = Note(
                fields={
                    self.Front.field_name: annotation.original_text,
                    self.Back.field_name: annotation.annotations,
                }
            )
            yield foreign_key, note
